'''
This module contains objects and methods for extracting
data from quantum chemistry program files
'''

from abc import ABC, abstractmethod
import numpy as np
import numpy.linalg as la
import xyz_py as xyzp
import numpy.typing as npt
import sys
import datetime

from . import string_tools as st
from .__version__ import __version__
from . import utils as ut

class QCStructure(ABC):
    '''
    Abstract Base Class (template) for Quantum Chemistry Structure classes
    '''

    def __init__(self, file_name, labels, coords):
        self.file_name = file_name
        self.labels = labels
        self.coords = coords
        self.n_atoms = len(labels)

        return

    @staticmethod
    def guess_from_file(file_name: str) -> 'QCCS':
        '''
        Attempts to figure out which class to use to extract shift data
        from a given quantum chemistry output file

        Parameters
        ----------
        file_name: str
            Name of file to examine

        Returns
        -------
        QCCS
            Quantum Chemistry Structure data object
        '''

        SUPPORTED_READERS: list[QCA] = [
            OrcaOutputStructure,
            GaussianLogStructure,
        ]

        data = None

        with open(file_name, 'r') as f:
            for line in f:
                for obj in SUPPORTED_READERS:
                    if obj.COMMON_STR in line:
                        # Load quantum chemical hyperfine data
                        data = obj.read(file_name)
                        break

        if data is None:
            sys.exit(f'Cannot find data in {file_name}')

        return data

    'string name of filetype'
    FILETYPE: str

    'string to look for in file which identifies type of file'
    COMMON_STR: str

    'String name of file which has been read'
    file_name: str

    'Number of atoms in system'
    n_atoms: int

    'Atomic labels, with indexing numbers'
    labels: npt.NDArray[np.str_]

    'Atomic coordinates (3xn_atoms)'
    coords: npt.NDArray

    @classmethod
    def read(cls, file_name: str):
        '''
        DO NOT EDIT THIS
        Wrapper for user implemented _read function which checks existence and
        type of required attributes
        '''

        instance = cls._read(file_name)

        attributes = [
            'FILETYPE',
            'COMMON_STR',
            'file_name',
            'n_atoms',
            'labels',
            'coords'
        ]

        for attribute in attributes:
            try:
                getattr(instance, attribute)
            except AttributeError:
                sys.exit(
                    'ERROR: Attribute {} is missing from {}'.format(
                        attribute, cls
                    )
                )

        return instance

    @classmethod
    @abstractmethod
    def _read(file_name: str):
        '''
        User implemented function which extracts data from QC files and
        creates object
        '''
        raise NotImplementedError


class GaussianLogStructure(QCStructure):
    '''
    Structure object for Gaussian LOG files
    '''
    FILETYPE = 'Gaussian LOG'

    COMMON_STR = 'Gaussian(R)'

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        labels, coords = read_gaussian_log_xyz(file_name)
        labels = np.array(xyzp.add_label_indices(labels))

        return cls(file_name, labels, coords)


class OrcaOutputStructure(QCStructure):
    '''
    Structure object for Orca Output files
    '''
    FILETYPE = 'Orca OUTPUT'

    COMMON_STR = '* O   R   C   A *'

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        old_labels, coords = read_orca5_output_xyz(file_name)
        old_labels = np.array(
            xyzp.add_label_indices(
                old_labels,
                style='sequential',
                start_index=0
            )
        )
        new_labels = np.array(
            xyzp.add_label_indices(
                xyzp.remove_label_indices(old_labels)
            )
        )

        return cls(file_name, new_labels, coords)


class QCCS(ABC):
    '''
    Abstract Base Class (template) for Quantum Chemistry Chemical Shielding
    classes.
    '''

    def __init__(self, file_name, labels, coords, cs_iso, cs_aniso, cs_units):
        '''
        Constructor
        '''

        self.file_name = file_name
        self.labels = labels
        self.coords = coords
        self.n_atoms = len(labels)
        self.cs_iso = cs_iso
        self.cs_aniso = cs_aniso
        self.cs_units = cs_units

        return

    @staticmethod
    def guess_from_file(file_name: str) -> 'QCCS':
        '''
        Attempts to figure out which class to use to extract shift data
        from a given quantum chemistry output file

        Parameters
        ----------
        file_name: str
            Name of file to examine

        Returns
        -------
        QCCS
            Quantum Chemistry Chemical Shift data object
        '''

        SUPPORTED_CS_OBJ: list[QCA] = [
            OrcaOutputCS,
            OrcaPropertyCS,
            Gaussian09LogCS,
            Gaussian16LogCS
        ]

        data = None

        with open(file_name, 'r') as f:
            for line in f:
                for obj in SUPPORTED_CS_OBJ:
                    if obj.COMMON_STR in line:
                        # Load quantum chemical hyperfine data
                        data = obj.read(file_name)
                        break

        if data is None:
            sys.exit(f'Cannot find data in {file_name}')

        return data

    def __str__(self):
        '''
        String representation of class used for printing parsed information
        to output file
        '''

        string = ''

        string += st.title('Quantum Chemistry Chemical Shielding Data')

        string += 'Data was read from: {}\n'.format(self.file_name)

        string += 'As filetype: {}\n'.format(self.FILETYPE)

        string += st.subtitle('Coordinates (Å)')

        for label, coord in zip(self.labels, self.coords):
            string += '{:5}  {: 10.6f}  {: 10.6f}  {: 10.6f}\n'.format(
                label, *coord
            )

        string += st.subtitle('Isotropic Chemical Shielding ({})'.format(
            self.cs_units
        ))

        for label, val in self.cs_iso.items():
            string += '{:5} {: .6f}\n'.format(label, val)

        string += st.subtitle('Anisotropic Chemical Shielding ({})'.format(
            self.cs_units
        ))

        for label, val in self.cs_aniso.items():
            string += '{:5} {: .6f}\n'.format(label, val)

        return string

    'string name of filetype'
    FILETYPE: str

    'string to look for in file which identifies type of file'
    COMMON_STR: str

    'String name of file which has been read'
    file_name: str

    'Number of atoms in system'
    n_atoms: int

    'Atomic labels, with indexing numbers'
    labels: npt.NDArray[np.str_]

    'Atomic coordinates (3xn_atoms)'
    coords: npt.NDArray

    'Isotropic Chemical Shielding values'
    cs_iso: dict[str, float]

    'Anisotropic Chemical Shielding values'
    cs_aniso: dict[str, float]

    '''
    Units of Isotropic Chemical Shielding (cs)
    '''
    cs_units: str

    @classmethod
    def read(cls, file_name: str):
        '''
        DO NOT EDIT THIS
        Wrapper for user implemented _read function which checks existence and
        type of required attributes
        '''

        instance = cls._read(file_name)

        attributes = [
            'FILETYPE',
            'COMMON_STR',
            'file_name',
            'n_atoms',
            'labels',
            'coords',
            'cs_iso',
            'cs_aniso',
            'cs_units'
        ]

        for attribute in attributes:
            try:
                getattr(instance, attribute)
            except AttributeError:
                sys.exit(
                    'ERROR: Attribute {} is missing from {}'.format(
                        attribute, cls
                    )
                )

        return instance

    @classmethod
    @abstractmethod
    def _read(file_name: str):
        '''
        User implemented function which extracts data from QC files and
        creates object
        '''
        raise NotImplementedError


class OrcaOutputCS(QCCS):
    '''
    Chemical Shielding object for Orca OUTPUT files
    '''
    FILETYPE = 'Orca OUTPUT'

    COMMON_STR = '* O   R   C   A *'

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        old_labels, coords = read_orca5_output_xyz(file_name)
        old_labels = np.array(
            xyzp.add_label_indices(
                old_labels,
                style='sequential',
                start_index=0
            )
        )
        cs_iso, cs_aniso = read_orca5_output_cs(file_name)

        new_labels = np.array(
            xyzp.add_label_indices(
                xyzp.remove_label_indices(old_labels)
            )
        )

        converter = {
            ol: nl
            for ol, nl in zip(old_labels, new_labels)
        }

        cs_iso = {
            converter[label]: val
            for label, val in cs_iso.items()
        }

        cs_aniso = {
            converter[label]: val
            for label, val in cs_aniso.items()
        }

        cs_units = 'ppm'

        return cls(file_name, new_labels, coords, cs_iso, cs_aniso, cs_units)


class OrcaPropertyCS(QCCS):
    '''
    Chemical Shielding object for Orca PROPERTY files
    '''
    FILETYPE = 'Orca PROPERTY'

    COMMON_STR = '!PROPERTIES!'

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        old_labels, coords = read_orca5_property_xyz(file_name)
        cs_iso, cs_aniso = read_orca5_property_cs(file_name)

        # Convert orca labelling 1-> natoms for all atoms
        # to 1-n_atoms per element
        new_labels = np.array(
            xyzp.add_label_indices(
                xyzp.remove_label_indices(old_labels)
            )
        )
        converter = {
            old: new
            for old, new in zip(old_labels, new_labels)
        }

        cs_iso = {
            converter[label]: value
            for label, value in cs_iso.items()
        }
        cs_aniso = {
            converter[label]: tensor
            for label, tensor in cs_aniso.items()
        }

        cs_units = 'ppm'

        return cls(file_name, new_labels, coords, cs_iso, cs_aniso, cs_units)


class Gaussian16LogCS(QCCS):
    '''
    Chemical Shielding object for Gaussian LOG files
    '''
    FILETYPE = 'Gaussian LOG'

    COMMON_STR = 'Gaussian(R) 16 program'

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        labels, coords = read_gaussian_log_xyz(file_name)
        labels = np.array(xyzp.add_label_indices(labels))
        cs_iso, cs_aniso = read_gaussian16_log_cs(file_name)

        cs_units = 'ppm'

        return cls(file_name, labels, coords, cs_iso, cs_aniso, cs_units)


class Gaussian09LogCS(QCCS):
    '''
    Chemical Shielding object for Gaussian LOG files
    '''
    FILETYPE = 'Gaussian LOG'

    COMMON_STR = 'Gaussian(R) 09 program'

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        labels, coords = read_gaussian_log_xyz(file_name)
        labels = np.array(xyzp.add_label_indices(labels))
        cs_iso, cs_aniso = read_gaussian09_log_cs(file_name)

        cs_units = 'ppm'

        return cls(file_name, labels, coords, cs_iso, cs_aniso, cs_units)

class QCSpin(ABC):
    """
    Abstract Base Class for Quantum Chemistry Spin Data classes.
    """

    def __init__(self, file_name: str, S: float, multiplicity: int):
        self.file_name = file_name
        self.S = S
        self.multiplicity = multiplicity

    @staticmethod
    def guess_from_file(file_name: str) -> 'QCSpin':
        SUPPORTED_SPIN_OBJS: list[type["QCSpin"]] = [
            GaussianLogSpin,
            OrcaSpin
        ]

        data = None
        with open(file_name, 'r') as f:
            for line in f:
                for obj in SUPPORTED_SPIN_OBJS:
                    if obj.COMMON_STR in line:
                        data = obj.read(file_name)
                        break
                if data is not None:
                    break
        if data is None:
            sys.exit(f"Cannot find spin data in {file_name}")

        return data

    FILETYPE: str
    COMMON_STR: str
    file_name: str
    S: float
    multiplicity: int | None

    @classmethod
    def read(cls, file_name: str) -> "QCSpin":
        instance = cls._read(file_name)
        for attribute in ["FILETYPE", "COMMON_STR", "file_name", "S", "multiplicity"]:
            try:
                getattr(instance, attribute)
            except AttributeError:
                sys.exit(f"Attribute {attribute} is missing from {cls}")
        return instance

    @classmethod
    @abstractmethod
    def _read(cls, file_name: str) -> "QCSpin":
        raise NotImplementedError

class GaussianLogSpin(QCSpin):
    """
    Spin object for Gaussian LOG files
    """
    FILETYPE = "Gaussian LOG"
    COMMON_STR = "Gaussian(R)"

    @classmethod
    def _read(cls, file_name: str) -> "GaussianLogSpin":
        multiplicity = read_gaussian_log_spin(file_name)
        S = (multiplicity - 1) / 2.0
        return cls(file_name, S, multiplicity)


class OrcaSpin(QCSpin):
    """
    Spin object for Orca OUTPUT files
    """
    FILETYPE = "Orca OUTPUT"
    COMMON_STR = "* O   R   C   A *"

    @classmethod
    def _read(cls, file_name: str) -> "OrcaSpin":
        S = read_orca_spin(file_name)
        multiplicity = int(2 * S + 1)
        return cls(file_name, S, multiplicity)


class QCA(ABC):
    '''
    Abstract Base Class (template) for Quantum Chemistry A Tensor (Hyperfine)
    Data classes.
    '''

    def __init__(self, file_name, labels, coords, a_iso, a_dip, a_units):
        '''
        Constructor
        '''

        self.file_name = file_name
        self.labels = labels
        self.coords = coords
        self.n_atoms = len(labels)
        self.a_iso = a_iso
        self.a_dip = a_dip
        self.a_units = a_units

        return

    @staticmethod
    def guess_from_file(file_name: str) -> 'QCA':
        '''
        Attempts to figure out which class to use to extract A tensor
        from a given quantum chemistry output file

        Parameters
        ----------
        file_name: str
            Name of file to examine

        Returns
        -------
        str {'Orca PROPERTY', 'Orca OUTPUT', 'Gaussian LOG'}
            String identifier of relevant QCA implementation
            which matches QCA.FILETYPE attribute
        '''

        SUPPORTED_A_OBJS: list[QCA] = [
            GaussianLogA,
            Orca5OutputA,
            Orca6OutputA,
            Orca5PropertyA
        ]

        with open(file_name, 'r') as f:
            for line in f:
                for obj in SUPPORTED_A_OBJS:
                    if obj.COMMON_STR in line:
                        # Load quantum chemical hyperfine data
                        data = obj.read(file_name)
                        break

        if data is None:
            sys.exit(f'Cannot find data in {file_name}')

        return data

    def save_to_csv(self, file_name: str = 'dft_hyperfines.csv',
                    verbose: bool = True, comment: str = '',
                    delimiter: str = ',') -> None:
        '''
        Saves Quantum Chemistry Hyperfine data to file

        Parameters
        ----------
        file_name: str
            File to which hyperfine data is written in CSV format
        verbose: bool, default True
            If True, echo filename to screen
        comment: str, optional
            Additional comment line WITH comment character
        delimiter: str, default ','
            Delimiter used in output CSV file
        '''

        # Save hyperfine data to file
        out = np.array(
            [
                '{}, {:.5f}, {:.5f}, {:.5f}, {:.5f}, {:.5f}, {:.5f}, {:.5f}'.format( # noqa
                    label, iso, *tensor[0, :], *tensor[1, 1:], tensor[2, 2]
                )
                for iso, (label, tensor) in zip(self.a_iso.values(), self.a_dip.items()) # noqa
            ]
        )

        _comments = (
            f'#This file was generated with SimpNMR v{__version__}'
            ' on {}\n'.format(
                datetime.datetime.now().strftime('%H:%M:%S %d-%m-%Y ')
            )
        )

        _comments += comment + '\n'

        header = f'atom_label, Aiso ({self.a_units}), Adip_xx ({self.a_units}), Adip_xy ({self.a_units}), Adip_xz ({self.a_units}), Adip_yy ({self.a_units}), Adip_yz ({self.a_units}), Adip_zz ({self.a_units})' # noqa

        # Save to file
        np.savetxt(
            file_name,
            out,
            delimiter=delimiter,
            header=header,
            fmt='%s',
            comments=_comments
        )

        if verbose:
            ut.cprint(
                f'\n Raw DFT Hyperfine data written to \n {file_name}\n',
                'cyan'
            )

        return

    def __str__(self):
        '''
        String representation of class used for printing parsed information
        to output file
        '''

        string = ''

        string += st.title('Quantum Chemistry Hyperfine Data')

        string += 'Data was read from: {}\n'.format(self.file_name)

        string += 'As filetype: {}\n'.format(self.FILETYPE)

        string += st.subtitle('Coordinates (Å)')

        for label, coord in zip(self.labels, self.coords):
            string += '{:5}  {: 10.6f}  {: 10.6f}  {: 10.6f}\n'.format(
                label, *coord
            )

        string += st.subtitle('Isotropic A values ({})'.format(
            self.a_units
        ))

        for label, val in self.a_iso.items():
            string += '{:5} {: .6f}\n'.format(label, val)

        string += st.subtitle('Anisotropic (dipolar) A Tensor ({})'.format(
            self.a_units
        ))

        for label, tensor in self.a_dip.items():
            string += '\n      {: .6f} {: .6f} {: .6f}\n'.format(*tensor[0])
            string += '{:5} {: .6f} {: .6f} {: .6f}\n'.format(
                label, *tensor[1]
            )
            string += '      {: .6f} {: .6f} {: .6f}\n'.format(*tensor[2])

        return string

    'string name of filetype'
    FILETYPE: str

    'string to look for in file which identifies type of file'
    COMMON_STR: str

    'String name of file which has been read'
    file_name: str

    'Number of atoms in system'
    n_atoms: int

    'Atomic labels, with indexing numbers'
    labels: npt.NDArray[np.str_]

    'Atomic coordinates (3xn_atoms)'
    coords: npt.NDArray

    'Isotropic Hyperfine coupling values'
    a_iso: dict[str, float]

    '''Anisotropic (dipolar) Hyperfine coupling tensors (traceless)
    keys are string label with index number, values are (3x3) arrays'''
    a_dip: dict[str, npt.NDArray]

    '''
    Units of A_iso and A_dip
    '''
    a_units: str

    @classmethod
    def read(cls, file_name: str):
        '''
        DO NOT EDIT THIS
        Wrapper for user implemented _read function which checks existence and
        type of required attributes
        '''

        instance = cls._read(file_name)

        attributes = [
            'FILETYPE',
            'COMMON_STR',
            'file_name',
            'n_atoms',
            'labels',
            'coords',
            'a_iso',
            'a_dip',
            'a_units'
        ]

        for attribute in attributes:
            try:
                getattr(instance, attribute)
            except AttributeError:
                sys.exit(
                    'ERROR: Attribute {} is missing from {}'.format(
                        attribute, cls
                    )
                )

        return instance

    @classmethod
    @abstractmethod
    def _read(file_name: str):
        '''
        User implemented function which extracts data from QC files and
        creates class
        '''
        raise NotImplementedError


class GaussianLogA(QCA):
    '''
    A QCA object for Gaussian LOG files
    '''

    FILETYPE = 'Gaussian LOG'

    COMMON_STR = 'Gaussian(R)'

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        labels, coords = read_gaussian_log_xyz(file_name)
        labels = np.array(xyzp.add_label_indices(labels))
        a_iso_raw, a_dip_raw = read_gaussian_log_a_tensors(file_name)

        mult = read_gaussian_log_spin(file_name)
        n_unpaired = mult - 1

        # Convert to dict
        a_iso = {
            label: val
            for label, val in zip(labels, a_iso_raw)
        }

        # Convert to dict
        # and normalise by number of unpaired electrons
        a_dip = {
            label: tensor * 1. / n_unpaired
            for label, tensor in zip(labels, a_dip_raw)
        }

        a_units = 'MHz'

        return cls(file_name, labels, coords, a_iso, a_dip, a_units)


def read_gaussian_log_xyz(file_name: str) -> tuple[
        npt.NDArray[np.str_], npt.NDArray]:
    '''
    Read Gaussian .log file to extract atomic labels and coordinates

    Parameters
    ----------
        f_name : str
            Name of log file

    Returns:
    --------
        np.ndarray[str]
            List of atomic labels
        np.ndarray[float]
            (n_atom,3) array of atomic coordinates
    '''

    # Read number of atoms
    with open(file_name, 'r') as f:
        for line in f:
            if 'NAtoms=' in line:
                spl_line = line.split()
                n_atoms = int(spl_line[spl_line.index('NAtoms=') + 1])
                break

    # Get coordinates
    headers = ['Standard orientation:', 'Input orientation:']
    with open(file_name, 'r') as f:
        for line in f:
            if any([he in line for he in headers]):
                coords = []
                a_nums = []

                # Skip header
                for _ in range(4):
                    line = next(f)

                for _ in range(n_atoms):
                    line = next(f)
                    coords.append([float(coord) for coord in line.split()[3:]])
                    a_nums.append(int(line.split()[1]))

    f.close()

    # Convert atomic numbers to atomic labels
    labels = xyzp.num_to_lab(a_nums)

    labels = np.asarray(labels)
    coords = np.asarray(coords)

    return labels, coords


def read_gaussian_log_spin(file_name: str) -> int: # noqa
    '''
    Read Gaussian .log file to extract spin multiplicity (2S+1)

    Parameters
    ----------
        f_name : str
            Name of log file

    Returns:
    --------
        int
            spin multiplicity 2S+1
    '''

    # Read number of atoms
    with open(file_name, 'r') as f:
        for line in f:
            if 'Multiplicity =' in line:
                mult = int(line.split()[-1])

    return mult


def read_gaussian_log_a_tensors(file_name: str) -> tuple[
        npt.NDArray, npt.NDArray]:
    '''
    Extracts A tensor for each atom from Gaussian log file

    Parameters
    ----------
    file_name: str
        Name of log file

    Returns
    -------
    np.ndarray[float]
        A_iso values as n_atomsx1 array in MHz
    np.ndarray[float]
        A_dip tensors as n_atomsx3x3 array in MHz
    '''

    # Read number of atoms
    with open(file_name, 'r') as f:
        for line in f:
            if 'NAtoms=' in line:
                spl_line = line.split()
                n_atoms = int(spl_line[spl_line.index('NAtoms=') + 1])
                break

    a_iso = np.zeros(n_atoms)

    # Read isotropic part
    with open(file_name, 'r') as f:
        for line in f:
            if 'Isotropic Fermi Contact Couplings' in line:
                line = next(f)
                for it in range(n_atoms):
                    line = next(f)
                    a_iso[it] = float(line.split()[3])  # MHz

    a_dip = np.zeros([n_atoms, 3, 3])
    # Read traceless tensor as eigenvalues and eigenvectors
    track = 0
    with open(file_name, 'r') as f:
        for line in f:
            if 'Anisotropic Spin Dipole Couplings' in line:
                track += 1
                # Make sure in spin density part!
            if 'Anisotropic Spin Dipole Couplings' in line and track == 2:
                line = next(f)
                line = next(f)
                line = next(f)
                line = next(f)
                for it in range(n_atoms):
                    line = next(f)
                    val_1 = float(line.split()[2])  # MHz
                    vecs_1 = [float(val) for val in line.split()[-3:]]
                    line = next(f)
                    val_2 = float(line.split()[4])  # MHz
                    vecs_2 = [float(val) for val in line.split()[-3:]]
                    line = next(f)
                    val_3 = float(line.split()[2])  # MHz
                    vecs_3 = [float(val) for val in line.split()[-3:]]
                    vals = np.array([val_1, val_2, val_3])
                    vecs = np.array([vecs_1, vecs_2, vecs_3]).T

                    # Transform back to coordinate frame in MHz
                    a_dip[it, :, :] = vecs @ np.diag(vals) @ la.inv(vecs)
                    line = next(f)

    if track != 2:
        ut.cprint(
            (
                'Warning: Cannot find Dipolar Hyperfine Tensor in log file\n'
                ' Check prop=epr is in routecard!'
            ),
            'black_yellowbg'
        )

    return a_iso, a_dip


class Orca5OutputA(QCA):
    '''
    A Tensor object for Orca 5 OUTPUT files
    '''
    FILETYPE = 'Orca OUTPUT'

    COMMON_STR = "            '#,     ,#'  ##    ##  '#,     ,#' ,#      #,         ##   #,  ,#" # noqa

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        old_labels, coords = read_orca5_output_xyz(file_name)
        old_labels = np.array(
            xyzp.add_label_indices(
                old_labels,
                style='sequential',
                start_index=0
            )
        )
        a_iso, a_dip = read_orca5_output_a_tensors(file_name)

        new_labels = np.array(
            xyzp.add_label_indices(
                xyzp.remove_label_indices(old_labels)
            )
        )
        converter = {
            old: new
            for old, new in zip(old_labels, new_labels)
        }

        a_iso = {
            converter[label]: value
            for label, value in a_iso.items()
        }
        a_dip = {
            converter[label]: tensor
            for label, tensor in a_dip.items()
        }

        a_units = 'MHz'

        return cls(file_name, new_labels, coords, a_iso, a_dip, a_units)


class Orca6OutputA(QCA):
    '''
    A Tensor object for Orca 6 OUTPUT files
    '''
    FILETYPE = 'Orca OUTPUT'

    COMMON_STR = "            '#,     ,#'  ##    ##  '#,     ,#' ,#      #,     #,   #   #,  ,#" # noqa

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        old_labels, coords = read_orca5_output_xyz(file_name)
        old_labels = np.array(
            xyzp.add_label_indices(
                old_labels,
                style='sequential',
                start_index=0
            )
        )
        a_iso, a_dip = read_orca6_output_a_tensors(file_name)

        new_labels = np.array(
            xyzp.add_label_indices(
                xyzp.remove_label_indices(old_labels)
            )
        )
        converter = {
            old: new
            for old, new in zip(old_labels, new_labels)
        }

        a_iso = {
            converter[label]: value
            for label, value in a_iso.items()
        }
        a_dip = {
            converter[label]: tensor
            for label, tensor in a_dip.items()
        }

        a_units = 'MHz'

        return cls(file_name, new_labels, coords, a_iso, a_dip, a_units)


def read_orca5_output_xyz(file_name: str) -> tuple[npt.NDArray[np.str_], npt.NDArray]: # noqa
    '''
    Reads xyz coordinates from orca output file

    Takes final set of coordinates in file

    Parameters
    ----------
    file_name: str
        Name of orca output file

    Returns
    -------
    np.array:
        Atomic labels with no indexing numbers
    np.array:
        Atomic coordinates (xyz) as N_atoms,3 array in Angstrom
    '''

    labels, coords = [], []

    with open(file_name, 'r') as f:
        for line in f:
            if 'CARTESIAN COORDINATES (ANGSTROEM)' in line:
                labels, coords = [], []
                line = next(f)
                line = next(f)
                while len(line.split()):
                    labels.append(line.split()[0])
                    coords.append(line.split()[1:])
                    line = next(f)

    coords = [
        [float(trio[0]), float(trio[1]), float(trio[2])]
        for trio in coords
    ]

    labels = np.array(labels)
    coords = np.array(coords)

    return labels, coords


def read_orca6_output_a_tensors(file_name: str) -> tuple[
        dict[str, float], dict[str, npt.NDArray]]:
    '''
    Extracts A tensor for each atom from orca 5 output file\n
    as separate isotropic and dipolar parts only

    Parameters
    ----------
    file_name: str
        Name of log file

    Returns
    -------
    dict[str, float]
        A_iso for each atom (key) as float (val) in MHz
    dict[str, npt.NDArray}
        A-dip for each atom (key) as 3x3 array (val) in MHz
    '''

    # Find how many nuclei have been calculated
    with open(file_name, 'r') as f:
        for line in f:
            if 'ELECTRIC AND MAGNETIC HYPERFINE STRUCTURE' in line:
                n_calcd = int(line.split()[5][1:])

    a_iso = {}
    a_dip = {}

    # Read hyperfine data
    with open(file_name, 'r') as f:
        for line in f:
            if 'ELECTRIC AND MAGNETIC HYPERFINE STRUCTURE' in line:
                for _ in range(n_calcd):
                    while 'Nucleus' not in line:
                        line = next(f)
                    tmp = line.split()[1]
                    label = '{}{}'.format(
                        st.remove_numbers(tmp), st.remove_letters(tmp)
                    )
                    for _ in range(8):
                        line = next(f)

                    # Raw matrix in MHz
                    row_1 = [float(val) for val in line.split()]
                    line = next(f)
                    row_2 = [float(val) for val in line.split()]
                    line = next(f)
                    row_3 = [float(val) for val in line.split()]

                    full = np.array([
                        row_1, row_2, row_3
                    ])

                    a_iso[label] = 1 / 3 * np.trace(full)
                    a_dip[label] = full - np.eye(3) * a_iso[label]


    return a_iso, a_dip


def read_orca5_output_a_tensors(file_name: str) -> tuple[
        dict[str, float], dict[str, npt.NDArray]]:
    '''
    Extracts A tensor for each atom from orca 5 output file\n
    as separate isotropic and dipolar parts only

    Parameters
    ----------
    file_name: str
        Name of log file

    Returns
    -------
    dict[str, float]
        A_iso for each atom (key) as float (val) in MHz
    dict[str, npt.NDArray}
        A-dip for each atom (key) as 3x3 array (val) in MHz
    '''

    # Find how many nuclei have been calculated
    with open(file_name, 'r') as f:
        for line in f:
            if 'Number of nuclei for epr/nmr' in line:
                n_calcd = int(line.split()[-1])

    a_iso = {}
    a_dip = {}

    # Read hyperfine data
    with open(file_name, 'r') as f:
        for line in f:
            if 'ELECTRIC AND MAGNETIC HYPERFINE STRUCTURE' in line:
                line = next(f)
                line = next(f)
                line = next(f)
                for it in range(n_calcd):
                    line = next(f)
                    tmp = line.split()[1]
                    label = '{}{}'.format(
                        st.remove_numbers(tmp), st.remove_letters(tmp)
                    )
                    for _ in range(5):
                        line = next(f)

                    # Raw matrix in MHz
                    row_1 = [float(val) for val in line.split()]
                    line = next(f)
                    row_2 = [float(val) for val in line.split()]
                    line = next(f)
                    row_3 = [float(val) for val in line.split()]

                    for _ in range(5):
                        line = next(f)
                    a_iso[label] = float(line.split()[-1])

                    for _ in range(9):
                        line = next(f)

                    full = np.array([
                        row_1, row_2, row_3
                    ])

                    a_dip[label] = full - np.eye(3) * a_iso[label]

    return a_iso, a_dip


def read_orca5_output_cs(file_name: str) -> tuple[
        dict[str, float], dict[str, npt.NDArray]]:
    '''
    Extracts Chemical Shielding value for each atom from orca output file

    Parameters
    ----------
    file_name: str
        Name of log file

    Returns
    -------
    dict[str, float]
        Isotropic Chemical Shielding for each atom (key) as float (val) in ppm
    dict[str, float]
        Anisotropic Chemical Shielding for each atom (key) as float (val) in
        ppm
    '''

    cs_iso = {}
    cs_aniso = {}
    # Read Chemical Shielding data
    with open(file_name, 'r') as f:
        for line in f:
            if 'CHEMICAL SHIELDING SUMMARY (ppm)' in line:
                for _ in range(6):
                    line = next(f)

                while len(line.lstrip().rstrip()):
                    label = '{}{}'.format(
                        line.split()[1], int(line.split()[0])
                    )
                    cs_iso[label] = float(line.split()[2])
                    cs_aniso[label] = float(line.split()[3])
                    line = next(f)

    return cs_iso, cs_aniso


class Orca5PropertyA(QCA):
    '''
    A Tensor object for Orca PROPERTY files
    '''
    FILETYPE = 'Orca PROPERTY'

    COMMON_STR = "            '#,     ,#'  ##    ##  '#,     ,#' ,#      #,         ##   #,  ,#" # noqa

    @classmethod
    def _read(cls, file_name: str):

        # Read raw data
        old_labels, coords = read_orca5_property_xyz(file_name)
        a_iso, a_dip = read_orca5_property_a_tensors(file_name)

        # Convert orca labelling 1-> natoms for all atoms
        # to 1-n_atoms per element
        new_labels = np.array(
            xyzp.add_label_indices(
                xyzp.remove_label_indices(old_labels)
            )
        )
        converter = {
            old: new
            for old, new in zip(old_labels, new_labels)
        }

        a_iso = {
            converter[label]: value
            for label, value in a_iso.items()
        }
        a_dip = {
            converter[label]: tensor
            for label, tensor in a_dip.items()
        }

        a_units = 'MHz'

        return cls(file_name, new_labels, coords, a_iso, a_dip, a_units)


def read_orca5_property_xyz(file_name: str) -> tuple[npt.NDArray[np.str_], npt.NDArray]: # noqa
    '''
    Reads xyz coordinates from orca property file

    Takes final set of coordinates in file

    Parameters
    ----------
    file_name: str
        Name of orca property file

    Returns
    -------
    np.array:
        Atomic labels with indexing numbers
    np.array:
        Atomic coordinates (xyz) as N_atoms,3 array in Angstrom
    '''

    labels, coords = [], []

    with open(file_name, 'r') as f:
        for line in f:
            if '!GEOMETRY!' in line:
                line = next(f)
                n_atoms = int(line.split()[-1])
                for _ in range(2):
                    line = next(f)
                for _ in range(n_atoms):
                    line = next(f)
                    labels.append('{}{}'.format(
                        line.split()[1], line.split()[0]
                    ))
                    coords.append(line.split()[2:])

    coords = [
        [float(trio[0]), float(trio[1]), float(trio[2])]
        for trio in coords
    ]

    labels = np.array(labels)
    coords = np.array(coords)

    return labels, coords


def read_orca5_property_a_tensors(file_name: str) -> tuple[
        dict[str, float], dict[str, np.ndarray]]:
    '''
    Reads hyperfine coupling tensors from orca property file

    Parameters
    ----------
    file_name: str
        Name of orca file

    Returns
    -------
    dict[str, float]
        A_iso for each atom (key) as float (val) in MHz
    dict[str, npt.NDArray}
        A-dip for each atom (key) as 3x3 array (val) in MHz
    '''

    a_dip = {}
    a_iso = {}

    with open(file_name, 'r') as f:
        for line in f:
            if 'EPRNMR_ATensor' in line:
                while 'Number of stored nuclei' not in line:
                    line = next(f)
                n_calcd = int(line.split()[4])
                while 'Nucleus:' not in line:
                    line = next(f)
                for _ in range(n_calcd):
                    label = '{}{}'.format(
                        line.split()[2], line.split()[1]
                    )
                    for _ in range(6):
                        line = next(f)
                    # Raw values
                    row_1 = [float(val) for val in line.split()[1:]]
                    line = next(f)
                    row_2 = [float(val) for val in line.split()[1:]]
                    line = next(f)
                    row_3 = [float(val) for val in line.split()[1:]]
                    a_dip[label] = np.array(
                        [row_1, row_2, row_3]
                    )
                    for _ in range(9):
                        line = next(f)
                    # Isotropic value
                    a_iso[label] = float(line.split()[-1])
                    a_dip[label] -= np.eye(3) * a_iso[label]
                    line = next(f)

    return a_iso, a_dip


def read_orca5_property_cs(file_name: str) -> tuple[
        dict[str, float], dict[str, np.ndarray]]:
    '''
    Reads Chemical Shielding data from orca property file

    Parameters
    ----------
    file_name: str
        Name of orca file

    Returns
    -------
    dict[str, float]
        Isotropic Chemical Shielding for each atom (key) as float (val) in ppm
    dict[str, float]
        Anisotropic Chemical Shielding for each atom (key) as float (val) in
        ppm
    '''

    cs_iso = {}
    cs_aniso = {}

    with open(file_name, 'r') as f:
        for line in f:
            if 'EPRNMR_OrbitalShielding' in line:
                while 'Number of stored nuclei' not in line:
                    line = next(f)
                n_calcd = int(line.split()[4])
                while 'Nucleus:' not in line:
                    line = next(f)
                for _ in range(n_calcd):
                    label = '{}{}'.format(
                        line.split()[2], line.split()[1]
                    )
                    for _ in range(13):
                        line = next(f)
                    # Read eigenvalues and convert to Anisotropic CS
                    evals = np.array([float(val) for val in line.split()[1:]])
                    evals = sorted(evals)
                    cs_aniso[label] = evals[2] - (evals[0] + evals[1]) / 2.
                    line = next(f)
                    # Isotropic value
                    cs_iso[label] = float(line.split()[-1])
                    line = next(f)

    return cs_iso, cs_aniso


def read_gaussian09_log_cs(file_name):
    '''
    Reads Chemical Shielding data from Gaussian log file
    Parameters
    ----------
    file_name: str
        Name of gaussian log file

    Returns
    -------
    dict[str, float]
        Isotropic Chemical Shielding for each atom (key) as float (val) in ppm
    dict[str, float]
        Anisotropic Chemical Shielding for each atom (key) as float (val) in
        ppm
    '''

    cs_iso = {}
    cs_aniso = {}

    with open(file_name, 'r') as f:
        for line in f:
            if 'Magnetic shielding tensor (ppm)' in line:
                while 'Number of stored nuclei' not in line:
                    line = next(f)
                n_calcd = int(line.split()[4])
                while 'Nucleus:' not in line:
                    line = next(f)
                for _ in range(n_calcd):
                    label = '{}{}'.format(
                        line.split()[2], line.split()[1]
                    )
                    for _ in range(13):
                        line = next(f)
                    # Read eigenvalues and convert to Anisotropic CS
                    evals = np.array([float(val) for val in line.split()[1:]])
                    evals = sorted(evals)
                    cs_aniso[label] = evals[2] - (evals[0] + evals[1]) / 2.
                    line = next(f)
                    # Isotropic value
                    cs_iso[label] = float(line.split()[-1])
                    line = next(f)

    return cs_iso, cs_aniso


def read_gaussian16_log_cs(file_name):
    '''
    Reads Chemical Shielding data from Gaussian16 log file
    Parameters
    ----------
    file_name: str
        Name of gaussian log file

    Returns
    -------
    dict[str, float]
        Isotropic Chemical Shielding for each atom (key) as float (val) in ppm
    dict[str, float]
        Anisotropic Chemical Shielding for each atom (key) as float (val) in
        ppm
    '''

    cs_iso = {}
    cs_aniso = {}

    with open(file_name, 'r') as f:
        for line in f:
            if 'Isotropic =' in line:
                line = line.replace('=-', '= -')
                cs_iso['{}{:d}'.format(
                    line.split()[1],
                    int(line.split()[0])
                )] = float(line.split()[4])

    return cs_iso, cs_aniso


# def read_orca_susceptibility_method(file_name:str, section:str)->str:
#     '''
#     TODO:
#     Write description here

#     '''
#     if section == 'auto':
#         with open(file_name, 'r') as f:
#             for line in reversed(f):
#                 if f'QDPT WITH' in line:
#                     susc_method = split
#                     section = susc_method
#     else:
#         continue

#     return section


def read_orca_susceptibility(file_name: str, section: str) -> dict[float, np.ndarray]:
    '''
    Extracts temperature-dependent molar magnetic susceptibility tensors from an ORCA output file.

    Parameters
    ----------
    file_name : str
        Name of the ORCA output file.
    section : str
        Label of the QDPT section to read (for example, 'casscf' or 'nevpt2').

    Returns
    -------
    dict[float, np.ndarray]
        Dictionary mapping temperature in K to a 3x3 susceptibility tensor as a NumPy array.
    '''

    susceptibilities = {}

    with open(file_name, 'r') as f:
        for line in f:
            if f'QDPT WITH {section.upper()}' in line:
                while 'TEMPERATURE DEPENDENT MOLAR MAGNETIC SUSCEPTIBILITY TENSOR' not in line:  # noqa
                    line = next(f)
                # Move down until we reach the first temperature header line
                while 'TEMPERATURE/K' not in line:
                    line = next(f)
                while 'TEMPERATURE/K' in line:
                    _temp = float(line.split('TEMPERATURE/K:')[1])
                    line = next(f)
                    line = next(f)
                    # Read tensor
                    row_1 = [float(val) for val in line.split()]
                    line = next(f)
                    row_2 = [float(val) for val in line.split()]
                    line = next(f)
                    row_3 = [float(val) for val in line.split()]
                    susceptibilities[_temp] = np.array([row_1, row_2, row_3])
                    line = next(f)
                    line = next(f)

    return susceptibilities

def read_orca_spin(file_name: str) -> float:
    """
    Reads the spin quantum number S from an ORCA input-style line in the output,
    handling both:
        * xyz    charge mult
        * xyzfile charge mult filename.xyz

    Parameters
    ----------
    file_name : str
        Name of the ORCA output file.

    Returns
    -------
    float
        Spin quantum number S, derived from the spin multiplicity (2S+1).
    """
    spin = None

    with open(file_name, 'r') as f:
        for line in f:
            # Normalise whitespace to make matching robust
            compact = line.replace(' ', '').lower()
            if '*xyz' in compact:
                # Example lines:
                #   * xyz 0 2
                #   * xyzfile 0 2 ptbu3_opt_solv_optim.xyz
                tokens = line.split()
                # Collect all integer tokens (charge, multiplicity, etc.)
                int_tokens = []
                for tok in tokens:
                    stripped = tok.lstrip('+-')
                    if stripped.isdigit():
                        int_tokens.append(int(tok))
                if len(int_tokens) >= 2:
                    mult = int_tokens[1]  # second integer is multiplicity
                    spin = (mult - 1) / 2.0
                    break

    if spin is None:
        raise RuntimeError(
            f"Could not determine spin multiplicity from ORCA output '{file_name}'"
        )

    return spin

def read_orca_g_tensor(file_name: str, section: str) -> np.ndarray | None:
    '''
    Extracts the electronic g-tensor from the effective Hamiltonian section in an ORCA output file.

    Parameters
    ----------
    file_name : str
        Name of the ORCA output file.
    section : str
        Label of the QDPT section to read (for example, 'casscf' or 'nevpt2').

    Returns
    -------
    np.ndarray or None
        3x3 electronic g-tensor as a NumPy array if found, otherwise ``None``.
    '''

    g_tensor = None

    with open(file_name, "r") as f:
        for line in f:
            # Find the correct QDPT section
            if f"QDPT WITH {section.upper()}" in line:
                # Go down to the G-matrix header
                for line in f:
                    if "ELECTRONIC G-MATRIX FROM EFFECTIVE HAMILTONIAN" in line:
                        break
                # Find "g-matrix:"
                for line in f:
                    if "g-matrix:" in line:
                        # Next three lines are the rows of the tensor
                        row_1 = [float(val) for val in next(f).split()]
                        row_2 = [float(val) for val in next(f).split()]
                        row_3 = [float(val) for val in next(f).split()]
                        g_tensor = np.array([row_1, row_2, row_3])
                        break
                break

    return g_tensor

def read_eff_hamiltonian_tensor(file_name: str, section: str) -> np.ndarray | None:
    '''
    Extracts the raw electronic effective Hamiltonian tensor from an ORCA output file.

    Parameters
    ----------
    file_name : str
        Name of the ORCA output file.
    section : str
        Label of the QDPT section to read (for example, 'casscf' or 'nevpt2').

    Returns
    -------
    np.ndarray or None
        3x3 effective Hamiltonian tensor in cm-1 as a NumPy array if found, otherwise ``None``.
    '''

    eff_H_raw = None

    with open(file_name, "r") as f:
        for line in f:
            # Find the correct QDPT section
            if f"QDPT WITH {section.upper()}" in line:
                # Go down to the G-matrix header
                for line in f:
                    if "Effective Hamiltonian from projected relativistic states and relativistic energies:" in line:
                        break
                # Find "Raw matrix"
                for line in f:
                    if "Raw matrix (cm-1):" in line:
                        # Next three lines are the rows of the tensor
                        row_1 = [float(val) for val in next(f).split()]
                        row_2 = [float(val) for val in next(f).split()]
                        row_3 = [float(val) for val in next(f).split()]
                        eff_H_raw = np.array([row_1, row_2, row_3])
                        break
                break

    return eff_H_raw


from abc import ABC, abstractmethod
from scipy.optimize import least_squares, lsq_linear
from scipy.optimize._optimize import OptimizeResult
import numpy as np
from numpy.typing import NDArray
import numpy.linalg as la
import copy

from . import main
from . import utils as ut


class SusceptibilityModel(ABC):
    '''
    Abstract class on which all susceptibility fit classes are based
    '''

    def __init__(self, fit_vars: dict[str, float | str],
                 fix_vars: dict[str, float | str]):
        '''
        Set default values for mandatory attributes
        '''

        self.fit_vars = fit_vars
        self.fix_vars = fix_vars

        # Check all VARNAMES are provided in fit+fix
        input_names = [
            name for name in {**self.fit_vars, **self.fix_vars}.keys()
        ]

        if any([req_name not in input_names for req_name in self.VARNAMES]):
            raise ValueError(
                f'Missing fit/fix parameters in {self.NAME} Model'
            )

        # Final model parameter values
        self._final_var_values = {
            var: None
            for var in self.VARNAMES
        }
        # Standard deviation of each parameter
        self._fit_stdev = {
            var: None
            for var in self.fit_vars.keys()
        }

        # Fit status and temperature
        self._fit_status = False
        self._temperature = None

        # r2 and adjusted r2
        self._r2 = None
        self._adj_r2 = None

        # Residual
        self._mae = None
        # RMSE
        self._rmse = None

        return

    @property
    def fit_status(self) -> bool:
        'True if fit successful, else False'
        return self._fit_status

    @fit_status.setter
    def fit_status(self, value: bool):
        if isinstance(value, bool):
            self._fit_status = value
        else:
            raise TypeError
        return

    @property
    def temperature(self) -> float:
        'Temperature of fit (K)'
        return self._temperature

    @temperature.setter
    def temperature(self, value):
        if isinstance(value, (np.floating, float)):
            self._temperature = value
        else:
            raise TypeError
        return

    @property
    def final_var_values(self) -> float:
        'Final values of parameters, both fitted and fixed'
        return self._final_var_values

    @final_var_values.setter
    def final_var_values(self, value: dict):
        if isinstance(value, dict):
            self._final_var_values = value
        else:
            raise TypeError
        return

    @property
    def fit_stdev(self) -> float:
        'Standard deviation on fitted parameters, from fitting routine'
        return self._fit_stdev

    @fit_stdev.setter
    def fit_stdev(self, value: dict):
        if isinstance(value, dict):
            self._fit_stdev = value
        else:
            raise TypeError
        return

    @property
    def fix_vars(self) -> dict[str, float]:
        '''
        Parameters to fix (i.e. not fit)
        keys are names in VARNAMES, values are values
        '''
        return self._fix_vars

    @fix_vars.setter
    def fix_vars(self, value: dict):
        if isinstance(value, dict):
            unknown = [key for key in value.keys() if key not in self.VARNAMES]
            if any(unknown):
                raise KeyError(
                    f'Unknown variable names {unknown} provided to fix'
                )
            self._fix_vars = value
        else:
            raise TypeError('fix must be dictionary')
        return

    @property
    def fit_vars(self) -> dict[str, float]:
        '''
        Parameters to fit
        keys are names in VARNAMES, values are values
        '''
        return self._fit_vars

    @fit_vars.setter
    def fit_vars(self, value: dict):
        if isinstance(value, dict):
            unknown = [key for key in value.keys() if key not in self.VARNAMES]
            if any(unknown):
                raise KeyError(
                    f'Unknown variable names {unknown} provided to fix'
                )
            self._fit_vars = value
        else:
            raise TypeError('Fit must be dictionary')

        # Reset final model parameter values
        self._final_var_values = {
            var: None
            for var in self.VARNAMES
        }
        # Reset standard deviation of each parameter
        self._fit_stdev = {
            var: None
            for var in self.fit_vars.keys()
        }
        return

    @property
    def r2(self) -> float:
        'r2 of fit'
        return self._r2

    @r2.setter
    def r2(self, value):
        if isinstance(value, (np.floating, float)):
            self._r2 = value
        else:
            raise TypeError
        return

    @property
    def adj_r2(self) -> float:
        'Adjusted r2 of fit'
        return self._adj_r2

    @adj_r2.setter
    def adj_r2(self, value):
        if isinstance(value, (np.floating, float)):
            self._adj_r2 = value
        else:
            raise TypeError
        return

    @property
    def mae(self) -> float:
        'Mean absolute error (MAE) of fit'
        return self._mae

    @mae.setter
    def mae(self, value):
        if isinstance(value, (np.floating, float)):
            self._mae = value
        else:
            raise TypeError
        return

    @property
    def rmse(self) -> float:
        'Root mean square error (RMSE) of fit'
        return self._rmse

    @rmse.setter
    def rmse(self, value):
        if isinstance(value, (np.floating, float)) or np.isnan(value):
            self._rmse = value
        else:
            raise TypeError
        return

    @property
    @abstractmethod
    def NAME() -> str:
        'string name of model'
        raise NotImplementedError

    @property
    @abstractmethod
    def VARNAMES() -> list[str]:
        'String names of parameters which can be fitted or fixed'
        raise NotImplementedError

    @property
    @abstractmethod
    def VARNAMES_MM() -> dict[str, str]:
        '''
        Mathmode (i.e. $$, latex ) versions of VARNAMES
        keys are strings from VARNAMES
        '''
        raise NotImplementedError

    @property
    @abstractmethod
    def UNITS_MM() -> dict[str, str]:
        '''
        Mathmode (i.e. $$, latex) versions of UNITS
        keys are strings from VARNAMES
        '''
        raise NotImplementedError

    @property
    @abstractmethod
    def BOUNDS() -> dict[str, list[float, float]]:
        '''
        Bounds for each parameter of model
        keys: parameter name
        values: [upper, lower]
        used by scipy least_squares
        '''
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def model(parameters: dict[str, float],
              nuclei: list[main.Nucleus]) -> dict[str, float]:
        '''
        Computes model function of paramagnetic chemical shift

        Parameters
        ----------
        parameters: dict[str, float]
            keys are VARNAMES, values are float values
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated

        Returns
        -------
        dict[str, float]
            keys are atom labels, values are shifts

        '''
        raise NotImplementedError

    def tosusceptibility(self) -> main.Susceptibility:
        '''
        Converts current model into main.Susceptibility object

        Parameters
        ----------
        None

        Returns
        -------
        main.Susceptibility
        '''

        tensor = self.totensor(self.final_var_values)

        susc = main.Susceptibility(tensor, self.temperature)

        return susc

    @staticmethod
    @abstractmethod
    def totensor(params: dict[str, float]) -> NDArray:
        '''
        Converts set of parameters for this model
        into a numpy array representation of the susceptibility tensor

        Parameters
        ----------
        params: dict[str, float]
            Keys are VARNAMES, values are float values

        Returns
        -------
        NDArray of floats
            Susceptibility tensor as numpy array
        '''
        raise NotImplementedError

    def residuals(self, parameters: dict[str, float],
                nuclei: list[main.Nucleus],
                al_to_para_shift: dict[str, float],
                average_labels: list[list[str]] = []) -> list[float]:
        '''
        Calculates difference between true susceptibility and trial
        susceptibility calculated using model

        Parameters
        ----------
        parameters: dict[str, float]
            parameters used in model function
            keys are VARNAMES, values are float values
        nuclei: list[Nuclei]
            Nuclei for which shifts will be calculated
        al_to_para_shift: dict[str, float]
            Atom label to experimental paramagnetic chemical shift
        average_labels: list[list[str]]
            Groups of atom labels whose shift is averaged prior to calculation
            of residual
        Returns
        -------
        list[float]
            vector of residuals, real, then imaginary
        '''

        trial_shifts = self.model(parameters, nuclei)

        # Initialize weights for all atom labels to 1.0
        weights = {lab: 1.0 for lab in trial_shifts.keys()}
        if average_labels:
            # For each group, compute the average shift and assign a weight factor
            # such that the overall contribution of the group is independent of its size
            for group in average_labels:
                group_average = np.mean([trial_shifts[lab] for lab in group])
                group_size = len(group)
                for lab in group:
                    trial_shifts[lab] = group_average
                    weights[lab] = np.sqrt(group_size)  # residuals will be divided by this
 
        # Compute residuals using uniform weighting for single signals and scaled weights for groups
        residuals = [
            (exp_shift - trial_shifts[atom_label]) / weights.get(atom_label, 1.0)
            for atom_label, exp_shift in al_to_para_shift.items()
        ]

        return residuals

    def residual_from_float_list(self, new_vals: list[float],
                                 fit_vars: dict[str, float],
                                 fix_vars: dict[str, float],
                                 nuclei: list[main.Nucleus],
                                 al_to_para_shift: dict[str, float],
                                 average_labels: list[list[str]] = []) -> list[float]: # noqa
        '''
        Wrapper for `residuals` method, takes new values from fitting routine
        which provides list[float], to construct new fit_vals dict, then
        runs `residuals` method.

        Parameters
        ----------
        new_vals: list[float]
            New values provided by fit routine, order matches fit_vars.keys()
        fit_vars: dict[str, float]
            Parameter to fit in model function
            keys are VARNAMES, values are float values
        fit_vars: dict[str, float]
            Parameter which remain fixed in model function
            keys are VARNAMES, values are float values
        nuclei: list[Nuclei]
            Nuclei for which shifts will be calculated
        al_to_para_shift: dict[str, float]
            Atom label to experimental paramagnetic chemical shift
        average_labels: list[list[str]]
            Groups of atom labels whose shift is averaged prior to calculation
            of residual
        Returns
        -------
        list[float]
            Residuals
        '''

        # Swap fit values for new values from fit routine
        new_fit_vars = {
            name: guess
            for guess, name in zip(new_vals, fit_vars.keys())
        }

        # And make combined dict of fit and fixed
        # variable names (keys) and values
        all_vars = {**fix_vars, **new_fit_vars}

        residuals = self.residuals(
            all_vars,
            nuclei,
            al_to_para_shift,
            average_labels=average_labels
        )

        return residuals

    def fit_to(self, molecule: main.Molecule, experiment: main.Experiment,
               verbose: bool = True,
               average_labels: list[list[str]] = []) -> None:
        '''
        Fits model to susceptibility data

        Parameters
        ----------
        molecule: main.Molecule
            Molecule to which a model will be fitted
        experiment: main.Experiment
            Experimental data object
        verbose: bool, default True
            If False, supresses terminal output
        average_labels: list[list[str]]
            Groups of atom labels whose shift is averaged prior to calculation
            of residual
        '''

        # Starting values
        guess = [val for val in self.fit_vars.values()]

        # Get bounds for variables to be fitted
        bounds = np.array([
            self.BOUNDS[name]
            for name in self.fit_vars.keys()
        ]).T

        # Chemical label to paramagnetic shift
        al_to_para_shift = {
            nuc.label: experiment[nuc.chem_label].shift - nuc.shift.dia
            for nuc in molecule.nuclei
        }

        curr_fit = least_squares(
            fun=self.residual_from_float_list,
            args=(
                self.fit_vars,
                self.fix_vars,
                molecule.nuclei,
                al_to_para_shift,
                average_labels
            ),
            x0=guess,
            bounds=bounds,
            jac='3-point'
        )

        self.temperature = experiment.temperature

        # Fitted parameters
        curr_fit_dict = {
            name: value
            for name, value in zip(self.fit_vars.keys(), curr_fit.x)
        }

        if curr_fit.status == 0:
            if verbose:
                ut.cprint(
                    f'\n Fit at {self.temperature} K failed - Too many iterations', # noqa
                    'black_yellowbg'
                )
            self.final_var_values = copy.deepcopy(curr_fit_dict)
            self.fit_stdev = {
                label: np.nan
                for label in self.fit_vars.keys()
            }
            self.fit_status = False
            self.mae = np.NaN
            self.rmse = np.NaN
            self.r2 = np.NaN
            self.adj_r2 = np.NaN
        else:
            # Calculate standard deviation error on the parameters
            stdev, _ = svd_stdev(curr_fit)

            # Standard deviation error on the parameters
            self.fit_stdev = {
                label: val
                for label, val in zip(self.fit_vars.keys(), stdev)
            }
            self.fit_status = True

            # Set fitted values
            self.final_var_values = copy.deepcopy(curr_fit_dict)
            # and fixed values
            for key, val in self.fix_vars.items():
                self.final_var_values[key] = val

            # R2
            self.mae = np.sum(np.abs(curr_fit.fun)) / len(curr_fit.fun)
            ss_res = np.sum(curr_fit.fun**2)
            self.rmse = np.sqrt(ss_res / len(curr_fit.fun))
            ecs = [
                al_to_para_shift[nuc.label]
                for nuc in molecule.nuclei
            ]
            ss_tot = np.sum((ecs - np.mean(ecs))**2)
            self.r2 = 1 - (ss_res / ss_tot)
            self.adj_r2 = 1 - (1 - self.r2) * (len(ecs) - 1) / (len(ecs) - len(self.fit_vars) - 1) # noqa

        return


class LinearSusceptibilityModel(SusceptibilityModel):
    def fit_to(self, molecule: main.Molecule, experiment: main.Experiment,
               verbose: bool = True,
               average_labels: list[list[str]] = []) -> None:
        '''
        Fits model to susceptibility data

        Parameters
        ----------
        molecule: main.Molecule
            Molecule to which a model will be fitted
        experiment: main.Experiment
            Experimental data object
        verbose: bool, default True
            If False, supresses terminal output
        average_labels: list[list[str]]
            Groups of atom labels whose shift is averaged prior to calculation
            of residual
        '''

        # Get bounds for variables to be fitted
        bounds = np.array([
            self.BOUNDS[name]
            for name in self.fit_vars.keys()
        ]).T

        curr_fit = lsq_linear(
            A=self.design_matrix(
                molecule.nuclei,
                self.fix_vars
            ),
            b=self.target_vector(
                molecule.nuclei,
                experiment,
                self.fix_vars
            ),
            bounds=bounds
        )

        self.temperature = experiment.temperature

        fit_var_names = [
            name for name in self.VARNAMES
            if name in self.fit_vars.keys()
        ]

        # Fitted parameters
        curr_fit_dict = {
            name: value
            for name, value in zip(fit_var_names, curr_fit.x)
        }

        if curr_fit.status == 0:
            if verbose:
                ut.cprint(
                    f'\n Fit at {self.temperature} K failed - Too many iterations', # noqa
                    'black_yellowbg'
                )
            self.final_var_values = copy.deepcopy(curr_fit_dict)
            self.fit_stdev = {
                label: np.nan
                for label in self.fit_vars.keys()
            }
            self.fit_status = False
            self.rmse = np.NaN
            self.r2 = np.NaN
            self.adj_r2 = np.NaN
        else:
            # Calculate Jacobian, here equal to the design matrix
            curr_fit.jac = self.design_matrix(
                molecule.nuclei,
                self.fix_vars
            )

            # Calculate standard deviation error on the parameters
            stdev, _ = svd_stdev(curr_fit)

            # Standard deviation error on the parameters
            self.fit_stdev = {
                label: val
                for label, val in zip(self.fit_vars.keys(), stdev)
            }
            self.fit_status = True

            # Set fitted values
            self.final_var_values = copy.deepcopy(curr_fit_dict)
            # and fixed values
            for key, val in self.fix_vars.items():
                self.final_var_values[key] = val

            # R2
            ss_res = np.sum(curr_fit.fun**2)
            self.rmse = np.sqrt(ss_res / len(curr_fit.fun))
            ecs = [
                experiment[nuc.chem_label]
                for nuc in molecule.nuclei
            ]
            ss_tot = np.sum((ecs - np.mean(ecs))**2)
            self.r2 = 1 - (ss_res / ss_tot)
            self.adj_r2 = 1 - (1 - self.r2) * (len(ecs) - 1) / (len(ecs) - len(self.fit_vars) - 1) # noqa

        return

    @staticmethod
    @abstractmethod
    def design_matrix(nuclei: list[main.Nucleus],
                      fix_vars: dict[str, float]):
        '''
        Computes model function of paramagnetic chemical shift

        Parameters
        ----------
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated
        fix_vars: dict[str, float]:
            Parameters to fix (i.e. not fit)
            keys are names in VARNAMES, values are float values

        Returns
        -------
        ndarray of floats
            Design matrix A in Ax=b problem
        '''

        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def target_vector(nuclei: list[main.Nucleus],
                      experiment: main.Experiment, fix_vars: dict[str, float]):
        '''
        Computes model function of paramagnetic chemical shift

        Parameters
        ----------
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated
        experiment: main.Experiment
            Experimental data object
        fix_vars: dict[str, float]:
            Parameters to fix (i.e. not fit)
            keys are names in VARNAMES, values are float values

        Returns
        -------
        ndarray of floats
            Target vector b in Ax=b problem
        '''

        raise NotImplementedError


class SplitFitter(SusceptibilityModel):

    NAME = 'Split Isotropic and Anisotropic Components of Susceptibility'

    VARNAMES = [
        'iso',
        'dxx',
        'dyy',
        'dxy',
        'dxz',
        'dyz'
    ]

    VARNAMES_MM = {
        'iso': r'$\chi_\mathregular{iso}$',
        'dxx': r'$\Delta\chi_{xx}$',
        'dyy': r'$\Delta\chi_{yy}$',
        'dxy': r'$\Delta\chi_{xy}$',
        'dxz': r'$\Delta\chi_{xz}$',
        'dyz': r'$\Delta\chi_{yz}$'
    }

    UNITS_MM = {
        'iso': r'Å$^3$',
        'dxx': r'Å$^3$',
        'dyy': r'Å$^3$',
        'dxy': r'Å$^3$',
        'dxz': r'Å$^3$',
        'dyz': r'Å$^3$'
    }

    BOUNDS = {
        'iso': [0., np.inf],
        'dxx': [-np.inf, np.inf],
        'dyy': [-np.inf, np.inf],
        'dxy': [-np.inf, np.inf],
        'dxz': [-np.inf, np.inf],
        'dyz': [-np.inf, np.inf]
    }

    @staticmethod
    def model(parameters: dict[str, float],
              nuclei: list[main.Nucleus]) -> dict[str, float]:
        '''
        Computes model function fermi + psuedo contact shift

        Parameters
        ----------
        parameters: dict[str, float]
            keys are VARNAMES, values are float values
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated

        Returns
        -------
        dict[str, float]
            keys are nucleus.label, values are paramagnetic shifts

        '''

        delta_params = copy.deepcopy(parameters)
        tnsr = SplitFitter.totensor(delta_params)

        shifts = {
            nuc.label: 1. / 3. * np.trace(tnsr @ nuc.A.tensor)
            for nuc in nuclei
        }

        return shifts

    @staticmethod
    def totensor(params: dict[str, float]) -> NDArray:
        '''
        Converts set of parameters for this model
        into a numpy array representation of the susceptibility tensor

        Parameters
        ----------
        params: dict[str, float]
            Keys are VARNAMES, values are float values

        Returns
        -------
        NDArray of floats
            Susceptibility tensor as numpy array
        '''

        tensor = np.array(
            [
                [
                    params['dxx'], params['dxy'], params['dxz']
                ],
                [
                    params['dxy'], params['dyy'], params['dyz']
                ],
                [
                    params['dxz'], params['dyz'], -params['dxx'] - params['dyy'] # noqa
                ]
            ]
        )
        tensor += np.eye(3) * params['iso']

        return tensor

class IsoAxRhoFitter(SusceptibilityModel):

    NAME = 'Isotropic, Axial, and Rhombic over Axial Components of Susceptibility'

    VARNAMES = [
        'iso',
        'ax',
        'rho_over_ax',
    ]

    VARNAMES_MM = {
        'iso': r'$\chi_\mathregular{iso}$',
        'ax': r'$\chi_\mathregular{ax}$',
        'rho_over_ax': r'$\chi_\mathregular{rho} / \chi_\mathregular{ax}$',
    }

    UNITS_MM = {
        'iso': r'Å$^3$',
        'ax': r'Å$^3$',
        'rho_over_ax': '', # need to solve the problem with units
    }

    BOUNDS = {
        'iso': [0., np.inf],
        'ax': [-np.inf, np.inf],
        'rho_over_ax': [0.0, 1/3],
    }

    @staticmethod
    def model(parameters: dict[str, float],
              nuclei: list[main.Nucleus]) -> dict[str, float]:
        '''
        Computes model function fermi + psuedo contact shift

        Parameters
        ----------
        parameters: dict[str, float]
            keys are VARNAMES, values are float values
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated

        Returns
        -------
        dict[str, float]
            keys are nucleus.label, values are paramagnetic shifts

        '''

        delta_params = copy.deepcopy(parameters)
        tnsr = IsoAxRhoFitter.totensor(delta_params)

        shifts = {
            nuc.label: 1. / 3. * np.trace(tnsr @ nuc.A.tensor)
            for nuc in nuclei
        }

        return shifts

    @staticmethod
    def totensor(params: dict[str, float]) -> NDArray:
        '''
        Converts set of parameters for this model
        into a numpy array representation of the susceptibility tensor

        Parameters
        ----------
        params: dict[str, float]
            Keys are VARNAMES, values are float values

        Returns
        -------
        NDArray of floats
            Susceptibility tensor as numpy array
        '''

        tensor = np.array(
            [
                [
                    -params['ax']/3 + params['rho_over_ax'] * params['ax'], 0.0, 0.0
                ],
                [
                    0.0, -params['ax']/3 - params['rho_over_ax'] * params['ax'], 0.0
                ],
                [
                    0.0, 0.0, 2/3 * params['ax']
                ]
            ]
        )
        tensor += np.eye(3) * params['iso']

        return tensor

class EigenFitter(SusceptibilityModel):

    NAME = 'Eigenvalue model of Susceptibility'

    VARNAMES = [
        'x',
        'y',
        'z'
    ]

    VARNAMES_MM = {
        'x': r'$\chi_\mathregular{x}$',
        'y': r'$\chi_\mathregular{y}$',
        'z': r'$\chi_\mathregular{z}$'
    }

    UNITS_MM = {
        'x': r'Å$^3$',
        'y': r'Å$^3$',
        'z': r'Å$^3$'
    }

    BOUNDS = {
        'x': [-np.inf, np.inf],
        'y': [-np.inf, np.inf],
        'z': [-np.inf, np.inf]
    }

    @staticmethod
    def model(parameters: dict[str, float],
              nuclei: list[main.Nucleus]) -> dict[str, float]:
        '''
        Computes model function fermi + (axial)psuedo contact shift

        Parameters
        ----------
        parameters: dict[str, float]
            keys are VARNAMES, values are float values
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated

        Returns
        -------
        dict[str, float]
            keys are nucleus.label, values are paramagnetic shifts
        '''
        chix = parameters['x']
        chiy = parameters['y']
        chiz = parameters['z']

        tnsr = np.array([
            [chix, 0, 0],
            [0, chiy, 0],
            [0, 0, chiz],
        ])

        shifts = {
            nuc.label: 1. / 3. * np.trace(tnsr @ nuc.A.tensor)
            for nuc in nuclei
        }

        return shifts

    @staticmethod
    def totensor(params: dict[str, float]) -> NDArray:
        '''
        Converts set of parameters for this model
        into a numpy array representation of the susceptibility tensor

        Parameters
        ----------
        params: dict[str, float]
            Keys are VARNAMES, values are float values

        Returns
        -------
        NDArray of floats
            Susceptibility tensor as numpy array
        '''

        chix = params['x']
        chiy = params['y']
        chiz = params['z']

        tensor = np.array([
            [chix, 0, 0],
            [0, chiy, 0],
            [0, 0, chiz],
        ])

        return tensor


class IsoEigenFitter(SusceptibilityModel):

    NAME = 'Eigenvalue model of Susceptibility'

    VARNAMES = [
        'dxx',
        'dyy',
        'iso'
    ]

    VARNAMES_MM = {
        'dxx': r'$\Delta\chi_\mathregular{xx}$',
        'dyy': r'$\Delta\chi_\mathregular{yy}$',
        'iso': r'$\chi_\mathregular{iso}$'
    }

    UNITS_MM = {
        'dxx': r'Å$^3$',
        'dyy': r'Å$^3$',
        'iso': r'Å$^3$'
    }

    BOUNDS = {
        'dxx': [-np.inf, np.inf],
        'dyy': [-np.inf, np.inf],
        'iso': [0, np.inf]
    }

    @staticmethod
    def model(parameters: dict[str, float],
              nuclei: list[main.Nucleus]) -> dict[str, float]:
        '''
        Computes model function:\n
        isotropic + axial\n
        where axial contribution is diagonal and assumed that A is in\n
        eigenframe of chi\n

        Parameters
        ----------
        parameters: dict[str, float]
            keys are VARNAMES, values are float values
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated

        Returns
        -------
        dict[str, float]
            keys are nucleus.label, values are paramagnetic shifts
        '''
        dxx = parameters['dxx']
        dyy = parameters['dyy']
        iso = parameters['iso']
        shifts = {
            nuc.label: nuc.A.iso * iso + nuc.shift.dia + 1. / 3. * dxx * (nuc.A.tensor[0, 0] - nuc.A.tensor[2, 2]) + 1. / 3. * dyy * (nuc.A.tensor[1, 1] - nuc.A.tensor[2, 2]) # noqa
            for nuc in nuclei
        }

        return shifts

    @staticmethod
    def totensor(params: dict[str, float]) -> NDArray:
        '''
        Converts set of parameters for this model
        into a numpy array representation of the susceptibility tensor

        Parameters
        ----------
        params: dict[str, float]
            Keys are VARNAMES, values are float values

        Returns
        -------
        NDArray of floats
            Susceptibility tensor as numpy array
        '''

        dxx = params['dxx']
        dyy = params['dyy']
        iso = params['iso']

        tensor = np.array([
            [dxx + iso, 0, 0],
            [0, dyy + iso, 0],
            [0, 0, -(dxx + dyy) + iso],
        ])

        return tensor

class FullSuscFitter(SusceptibilityModel):

    NAME = 'Full Susceptibility'

    VARNAMES = [
        'xx',
        'xy',
        'xz',
        'yy',
        'yz',
        'zz'
    ]

    VARNAMES_MM = {
        'xx': r'$\chi_{xx}$',
        'yy': r'$\chi_{yy}$',
        'zz': r'$\chi_{zz}$',
        'xy': r'$\chi_{xy}$',
        'xz': r'$\chi_{xz}$',
        'yz': r'$\chi_{yz}$'
    }

    UNITS_MM = {
        'xx': r'Å$^3$',
        'yy': r'Å$^3$',
        'zz': r'Å$^3$',
        'xy': r'Å$^3$',
        'xz': r'Å$^3$',
        'yz': r'Å$^3$',
    }

    BOUNDS = {
        'xx': [-np.inf, np.inf],
        'yy': [-np.inf, np.inf],
        'zz': [-np.inf, np.inf],
        'xy': [-np.inf, np.inf],
        'xz': [-np.inf, np.inf],
        'yz': [-np.inf, np.inf]
    }

    @staticmethod
    def model(parameters: dict[str, float],
              nuclei: list[main.Nucleus]) -> dict[str, float]:
        '''
        Computes model function of hyperfine chemical shift
        i.e. pseudocontact + Fermi

        Parameters
        ----------
        parameters: dict[str, float]
            keys are VARNAMES, values are float values
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated

        Returns
        -------
        dict[str, float]
            keys are nucleus.label, values are paramagnetic shifts
        '''

        tnsr = FullSuscFitter.totensor(parameters)

        shifts = {
            nuc.label: 1. / 3. * np.trace(tnsr @ nuc.A.tensor)
            for nuc in nuclei
        }

        return shifts

    @staticmethod
    def design_matrix(nuclei: list[main.Nucleus],
                      fix_vars: dict[str, float]):
        '''
        Computes model function of paramagnetic chemical shift

        Parameters
        ----------
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated
        fix_vars: dict[str, float]:
            Parameters to fix (i.e. not fit)
            keys are names in VARNAMES, values are float values

        Returns
        -------
        ndarray of floats
            Design matrix A in Ax=b problem
        '''
        to_pop = {
            'xx': 0,
            'xy': 1,
            'xz': 2,
            'yy': 3,
            'yz': 4,
            'zz': 5
        }

        design = []

        for nuc in nuclei:
            _vec = [
                nuc.A.tensor[0, 0],
                nuc.A.tensor[1, 0] + nuc.A.tensor[0, 1],
                nuc.A.tensor[0, 2] + nuc.A.tensor[2, 0],
                nuc.A.tensor[1, 1],
                nuc.A.tensor[2, 1] + nuc.A.tensor[1, 2],
                nuc.A.tensor[2, 2]
            ]

            for var in fix_vars.keys():
                _vec.pop(to_pop[var])
            design.append(_vec)

        design = 1. / 3. * np.asarray(design)

        return design

    @staticmethod
    def target_vector(nuclei: list[main.Nucleus],
                      experiment: main.Experiment, fix_vars: dict[str, float]):
        '''
        Computes model function of paramagnetic chemical shift

        Parameters
        ----------
        nuclei: list[main.Nucleus]
            Nuclei for which shifts will be calculated
        experiment: main.Experiment
            Experimental data object
        fix_vars: dict[str, float]:
            Parameters to fix (i.e. not fit)
            keys are names in VARNAMES, values are float values

        Returns
        -------
        ndarray of floats
            Target vector b in Ax=b problem
        '''

        target = []

        for nuc in nuclei:
            _tgt = experiment[nuc.chem_label]
            to_subtract = {
                'xx': nuc.A.tensor[0, 0],
                'xy': nuc.A.tensor[1, 0] + nuc.A.tensor[0, 1],
                'xz': nuc.A.tensor[0, 2] + nuc.A.tensor[2, 0],
                'yy': nuc.A.tensor[1, 1],
                'yz': nuc.A.tensor[2, 1] + nuc.A.tensor[1, 2],
                'zz': nuc.A.tensor[2, 2]
            }
            for key, val in fix_vars.items():
                _tgt -= 1. / 3. * to_subtract[key] * val

            _tgt -= nuc.shift.dia
            target.append(_tgt)

        target = np.asarray(target)

        return target


def svd_stdev(curr_fit: OptimizeResult) -> tuple[list[float], list[bool]]:
    '''
    Calculates standard deviation of fit-parameters given output of scipy
    least_squares.

    Uses SVD of jacobian to check for singlular values equal to zero.
    Singular values equal to zero are discarded, so the corresponding
    parameter has a standard deviation which cannot be computed and is
    instead set to zero
    (i.e. they are meaningless)

    Parameters
    ----------
    curr_fit: OptimizeResult
        Result object from scipy.optimise.least_squares

    Returns
    -------
    list[float]
        Standard deviation on parameters
    list[bool]
        bool for each parameter, if False, then standard deviation cannot be
        calculated (and is set to zero)
    '''

    # SVD of jacobian
    _, s, VT = la.svd(curr_fit.jac, full_matrices=False)
    # Zero threshold as multiple of machine precision
    threshold = np.finfo(float).eps * max(curr_fit.jac.shape) * s[0]
    # Find singular values = 0.
    nonzero_sing = s > threshold
    # Truncate to remove these values
    s = s[nonzero_sing]
    VT = VT[:s.size]
    # Calculate covariance of each parameter using truncated arrays
    pcov = VT.T / s**2 @ VT
    # Scale by reduced chi**2 to remove influence of input sigma (if present)
    # and just obtain standard deviation of fit
    chi2dof = np.sum(curr_fit.fun**2)
    chi2dof /= (curr_fit.fun.size - curr_fit.x.size)
    pcov *= chi2dof
    stdev = np.sqrt(np.diag(pcov))

    no_stdev = stdev > threshold

    if sum(nonzero_sing) == len(nonzero_sing):
        no_stdev = [True] * len(nonzero_sing)

    return stdev, no_stdev


def write_model_data(models: list[SusceptibilityModel],
                     file_name: str, verbose: bool = True) -> None:
    '''
    Writes parameters of a set of models to file.\n
    Assumes models are all of the same type, e.g. all IsoSuscFitter

    Parameters
    ----------
    models: list[SusceptibilityModel]
        Models, one per temperature, must all be same type
    file_name: str
        Name of output file
    verbose: bool, default True
        If True, file location is written to terminal

    Returns
    -------
    None
    '''
    f = open(file_name, 'w', encoding='utf-8')
    f.write(' {:^12} '.format('T'))

    # Fitted parameters
    for name in models[0].fit_vars.keys():
        f.write('{:^17} {:^17} '.format(name, name + '-s-dev'))

    # Fixed parameters
    for name in models[0].fix_vars.keys():
        f.write('{:^17} '.format(name))

    f.write('{:^12} '.format('r2'))
    f.write('{:^12} '.format('r2_adj'))

    f.write('\n')

    for model in models:
        if not model.fit_status:
            continue
        f.write('{:12.10f} '.format(model.temperature))

        for name in model.fit_vars.keys():
            f.write('{: 1.10E} {: 1.10E} '.format(
                model.final_var_values[name], model.fit_stdev[name]
            ))

        for value in model.fix_vars.values():
            f.write('{: 1.10E} '.format(value))

        f.write('{: 1.10E} '.format(model.r2))
        f.write('{: 1.10E} '.format(model.adj_r2))

        f.write('\n')

    if verbose:
        ut.cprint(
            f'\n Susceptibility Model parameters written to \n {file_name}\n',
            'cyan'
        )
    return



'''
This module contains objects and methods for working with paramagnetic nmr
spectra and susceptibility tensors
'''

import datetime
import pandas as pd
import scipy.constants as constants
import numpy as np

from . import utils as ut
from .__version__ import __version__
from . import main
from . import models


def save_susc(molecules: list[main.Molecule],
              file_name: str = 'susceptibility.csv',
              verbose: bool = True,
              susc_models: list[models.SusceptibilityModel] = [],
              susc_units: str = 'A3', delimiter: str = ',', comment: str = ''):
    '''
    Saves susceptibility tensors of a set of Molecules to csv file

    molecules: list[main.Molecule]
        List of molecules, each with an associated susceptibility object
    file_name: str, default 'susceptibility.csv'
        Name of file to save to
    verbose: bool, default True
        If True, print filename to screen
    susc_models: list[models.SusceptibilityModel], optional
        If included, parameter uncertainties will be added to output file
    delimiter: str, default ','
        Delimiter used in csv file
    susc_units: str, {'A3', 'A3 mol-1', 'cm3 mol-1', 'cm3'}
        Units to use for susceptibility
    comment: str, ''
        Additional comment line added to file - must begin with #
    '''

    if susc_units == 'A3':
        conv = 1.
        unit_label = r'Å^3'
    elif susc_units == 'A3 mol-1':
        conv = constants.Avogadro
        unit_label = r'Å^3 mol^-1'
    elif susc_units == 'cm3 ':
        conv = 1E-24
        unit_label = r'cm^3'
    elif susc_units == 'cm3 mol-1':
        conv = 1E-24 * constants.Avogadro / (4 * np.pi)
        unit_label = r'cm^3 mol^-1'

    # And write susceptibility tensor to csv
    out = {
        'Temperature (K)': [molecule.susc.temperature for molecule in molecules], # noqa
        f'chi_iso ({unit_label})': [molecule.susc.iso * conv for molecule in molecules], # noqa
        f'chi_iso-s-dev ({unit_label})': None,
        f'chi_xx ({unit_label})': [molecule.susc.tensor[0, 0] * conv for molecule in molecules], # noqa
        f'chi_xx-s-dev ({unit_label})': None,
        f'chi_xy ({unit_label})': [molecule.susc.tensor[0, 1] * conv for molecule in molecules], # noqa
        f'chi_xy-s-dev ({unit_label})': None,
        f'chi_xz ({unit_label})': [molecule.susc.tensor[0, 2] * conv for molecule in molecules], # noqa
        f'chi_xz-s-dev ({unit_label})': None,
        f'chi_yy ({unit_label})': [molecule.susc.tensor[1, 1] * conv for molecule in molecules], # noqa
        f'chi_yy-s-dev ({unit_label})': None,
        f'chi_yz ({unit_label})': [molecule.susc.tensor[1, 2] * conv for molecule in molecules], # noqa
        f'chi_yz-s-dev ({unit_label})': None,
        f'chi_zz ({unit_label})': [molecule.susc.tensor[2, 2] * conv for molecule in molecules], # noqa
        f'chi_zz-s-dev ({unit_label})': None,
        f'dchi_xx ({unit_label})': [molecule.susc.dtensor[0, 0] * conv for molecule in molecules], # noqa
        f'dchi_xx-s-dev ({unit_label})': None,
        f'dchi_xy ({unit_label})': [molecule.susc.dtensor[0, 1] * conv for molecule in molecules], # noqa
        f'dchi_xy-s-dev ({unit_label})': None,
        f'dchi_xz ({unit_label})': [molecule.susc.dtensor[0, 2] * conv for molecule in molecules], # noqa
        f'dchi_xz-s-dev ({unit_label})': None,
        f'dchi_yy ({unit_label})': [molecule.susc.dtensor[1, 1] * conv for molecule in molecules], # noqa
        f'dchi_yy-s-dev ({unit_label})': None,
        f'dchi_yz ({unit_label})': [molecule.susc.dtensor[1, 2] * conv for molecule in molecules], # noqa
        f'dchi_yz-s-dev ({unit_label})': None,
        f'dchi_zz ({unit_label})': [molecule.susc.dtensor[2, 2] * conv for molecule in molecules], # noqa
        f'dchi_zz-s-dev ({unit_label})': None,
        f'chi_x ({unit_label})': [molecule.susc.eigvals[0] * conv for molecule in molecules], # noqa
        f'chi_x-s-dev ({unit_label})': None,
        f'chi_y ({unit_label})': [molecule.susc.eigvals[1] * conv for molecule in molecules], # noqa
        f'chi_y-s-dev ({unit_label})': None,
        f'chi_z ({unit_label})': [molecule.susc.eigvals[2] * conv for molecule in molecules], # noqa
        f'chi_z-s-dev ({unit_label})': None,
        f'chi_ax ({unit_label})': [molecule.susc.axiality * conv for molecule in molecules], # noqa
        f'chi_ax-s-dev ({unit_label})': None,
        f'chi_rho ({unit_label})': [molecule.susc.rhombicity * conv for molecule in molecules], # noqa
        f'chi_rho-s-dev ({unit_label})': None,
        'alpha (degrees)': [molecule.susc.alpha for molecule in molecules], # noqa
        'alpha-s-dev (degrees)': None,
        'beta (degrees)': [molecule.susc.beta for molecule in molecules], # noqa
        'beta-s-dev (degrees)': None,
        'gamma (degrees)': [molecule.susc.gamma for molecule in molecules], # noqa
        'gamma-s-dev (degrees)': None,
        'r2 ()': None,
        'r2_adjusted ()': None,
        'MAE (ppm)': None,
        'RMSE (ppm)': None,
    }

    if len(susc_models):
        out['r2 ()'] = [model.r2 for model in susc_models]
        out['r2_adjusted ()'] = [model.adj_r2 for model in susc_models]
        out['MAE (ppm)'] = [model.mae for model in susc_models]
        out['RMSE (ppm)'] = [model.rmse for model in susc_models]
        for key in susc_models[0].fit_stdev:
            out[f'chi_{key}-s-dev ({unit_label})'] = [
                model.fit_stdev[key] * conv for model in susc_models
            ]

    to_pop = [key for key, value in out.items() if value is None]
    for pop in to_pop:
        out.pop(pop)

    df = pd.DataFrame(data=out)

    _comment = (
        f'#This file was generated with SimpNMR v{__version__}'
        ' at {}\n'.format(
            datetime.datetime.now().strftime('%H:%M:%S %d-%m-%Y ')
        )
    )

    if len(comment):
        if comment[0] != '#':
            comment = f'#{comment}'
        _comment += f'{comment}\n'

    with open(file_name, 'w') as _f:
        _f.write(_comment)

        df.to_csv(
            _f,
            sep=delimiter,
            header=True,
            float_format='%.5f',
            index=None
        )

    if verbose:
        ut.cprint(
            f'\n Susceptibility data written to \n {file_name}\n',
            'cyan'
        )

    return



'''
            SimpNMR

        Copyright (C) 2025

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

This script contains a program to split rate (ac and dc *_params.csv) files
'''

import argparse
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import os


def load_susceptibility_data(sources: dict[str, str],
                             index: int) -> list[dict[str, float]]:
    '''
    Loads susceptibility data from a range of sources

    Parameters
    ----------
    sources: dict[str, str]
        Keys are name of sources e.g. functional name\n
        Values are file in which data is stored

    Returns
    -------
    list[dict[str, float]]
    '''

    all_molecules = dict.fromkeys(sources, None)

    for source_name, source_file in sources.items():
        # Load quantum chemical hyperfine data
        _data = np.loadtxt(source_file, skiprows=1)

        all_molecules[source_name] = _data[index]

    return all_molecules


def plot_component(func_comps: dict[str, dict[str, float]], ylabel: str,
                   show: bool = True, save: bool = True,
                   fig: plt.Figure = None, ax: plt.Axes = None,
                   savename: str = 'hyperfines.png',
                   figure_title: str = 'Hyperfine coupling constants'):
    '''
    Plots hyperfine data for a set of different functionals

    Parameters
    ----------
    func_comps: dict[str, dict[str, float]]
        Outer dictionary keys are functional name\n
        values are susceptibility component
    ylabel: str
        ylabel of plot
    show: bool, default True
        Show plot
    show: bool, default True
        Save plot to `savename`
    figure_title: str, default 'Hyperfine coupling constants'
        Title of figure window

    Returns
    -------
    plt.Figure
        Figure
    plt.Axes
        Axes
    '''

    if None in [fig, ax]:
        fig, ax = plt.subplots(
            1,
            1,
            num=figure_title
        )

    ax.plot(func_comps.values(), lw=0, marker='x', color='k')

    ax.set_xticks(np.arange(len(func_comps)))
    ax.set_xticklabels(func_comps.keys(), rotation=45)

    ax.set_ylabel(ylabel)
    fig.tight_layout()

    if save:
        plt.savefig(f'{savename}', dpi=500)
    if show:
        plt.show()

    return fig, ax


def main():
    parser = argparse.ArgumentParser(
        description=(
            'This script allows you to plot multiple sets of Hyperfine data'
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument(
        'input_file',
        type=str,
        help='.csv file containing source file information'
    )

    parser.add_argument(
        '-w',
        '--window_append',
        type=str,
        help='Appends to window titles',
        default=''
    )

    uargs = parser.parse_args()

    # Load input file
    config = pd.read_csv(
        uargs.input_file,
        skip_blank_lines=True,
        skipinitialspace=True,
        comment='#'
    )

    # Make table each functional name and chi, r2a, and resid
    table = {}
    for name, folder in zip(config['name'], config['folder']):
        # Load susceptibility data
        _susc = pd.read_csv(
            os.path.join(folder, 'susceptibility_tensor.csv'),
            skip_blank_lines=True,
            skipinitialspace=True,
            comment='#'
        )

        table[name] = {
            'chi_iso': _susc['chi_iso (Å^3)'][0],
            'chi_ax': _susc['chi_ax (Å^3)'][0],
            'chi_rho': _susc['chi_rho (Å^3)'][0],
            'r2_adjusted': _susc['r2_adjusted ()'][0],
            'MAE': _susc['MAE (ppm)'][0]
        }

    # Isotropic parts
    plot_component(
        {name: val['chi_iso'] for name, val in table.items()},
        r'$\chi_\mathregular{iso} \mathregular{(\AA^{3})}$',
        show=False,
        figure_title='isotropic susceptibility ' + uargs.window_append
    )

    # Ax parts
    plot_component(
        {name: val['chi_ax'] for name, val in table.items()},
        r'$\Delta\chi_\mathregular{ax} \mathregular{(\AA^{3})}$',
        show=False,
        figure_title='axial susceptibility ' + uargs.window_append
    )

    # rhombic parts
    plot_component(
        {name: val['chi_rho'] for name, val in table.items()},
        r'$\Delta\chi_\mathregular{rho} \mathregular{(\AA^{3})}$',
        show=False,
        figure_title='rhombic susceptibility ' + uargs.window_append
    )

    plt.show()



"""
Module for rotation, alignment, and chi-frame transformations in PCS-related
coordinate mapping between NEVPT2 and DFT geometries. The module loads input
settings from a YAML configuration, reads susceptibility and hyperfine data,
and exposes `get_rotation_and_transformation` as its main public API.
"""
import re
import os
import glob
import datetime
import numpy as np
import numpy.linalg as la
import xyz_py as xyzp

from . import readers as rdrs
from . import utils as ut
from .inputs import PredictConfig
from .__version__ import __version__

def access_input_data():
    default_ymls = glob.glob(os.path.join(os.getcwd(), "*.yml"))

    if not default_ymls:
        raise FileNotFoundError("No .yml file found in current directory")

    INPUT_YML = os.environ.get("SIMPNMR_INPUT", default_ymls[0])

    cfg = PredictConfig.from_file(INPUT_YML)
    susc_path = cfg.susceptibility_file
    hfc_path = cfg.hyperfine_file

    # Temperatures come from YAML; we treat it as a single-element list for now
    temperature = cfg.susceptibility_temperatures

    # NEVPT2 coordinates
    nevpt2_labels, nevpt2_coords = rdrs.read_orca5_output_xyz(susc_path)

    # DFT coordinates
    qca = rdrs.QCA.guess_from_file(hfc_path)
    nevpt2_labels = qca.labels
    dft_coords = qca.coords

    # Susceptibility tensor
    chi_dict = rdrs.read_orca_susceptibility(susc_path, section="nevpt2")
    chiT = chi_dict[temperature[0]]

    return chiT, temperature, nevpt2_labels,nevpt2_coords, dft_coords


def get_rotation_and_transformation():
    """
    Compute and return both the rotation matrix (R) aligning NEVPT2 and DFT
    coordinate sets, and the final transformation matrix (trans_mat) using
    the susceptibility tensor.

    Returns
    -------
    tuple of numpy.ndarray
        R : (3, 3)
            Rotation matrix that aligns NEVPT2 geometry to DFT geometry.
        trans_mat : (3, 3)
            Transformation matrix used for PCS-related coordinate mapping.
    """

    chiT, temperature, _, nevpt2_coords, dft_coords = access_input_data()

    if np.allclose(nevpt2_coords, dft_coords, rtol=1e-6, atol=1e-8):
        return np.eye(3), np.eye(3)
    else:
        pass

    if len(nevpt2_coords) != len(dft_coords):
        raise ValueError

    # Compute rotation aligning NEVPT2 → DFT
    rot_mat, rmsd = xyzp.find_rotation(nevpt2_coords, dft_coords)

    # Temperature-normalised tensor
    chi = chiT / temperature[0]

    # Eigen-decomposition
    evals, evecs = la.eigh(chi)

    # Transformation matrix
    trans_mat = evecs.T @ rot_mat

    ut.cprint(
        f' Distinct Susceptibility and DFT geometries detected; applied rotational alignment (RMSD = {rmsd:.6f}). \n',
          'cyan'
          )
    
    # Need to add an additional functional to check if the HFC coords are in chi frame because it leads to the wrong prediction

    return rot_mat, trans_mat

def rotate_coords_to_chi_frame(file_path):

    chiT, _, nevpt2_labels, nevpt2_coords, _ = access_input_data()

    # Subtract isotropic component (trace)
    chiT_traceless = chiT - np.eye(3) * (np.trace(chiT) / 3.0)

    # Diagonalize matrix
    eigvals_traceless, eigvecs_traceless = la.eigh(chiT_traceless)

    idx = np.argsort(np.abs(eigvals_traceless))

    # Rotate eigenvectors so principal axis aligns with global Z
    eigvecs_sorted = eigvecs_traceless[:, idx]
    u = eigvecs_sorted[:, 2]
    z_axis = np.array([0.0, 0.0, 1.0])
    cross = np.cross(u, z_axis)
    if np.linalg.norm(cross) < 1e-8:
        R = np.eye(3)

    else:
        a = cross / np.linalg.norm(cross)
        theta = np.arccos(np.dot(u, z_axis))
        A = np.array([
            [0.0, -a[2], a[1]],
            [a[2], 0.0, -a[0]],
            [-a[1], a[0], 0.0]
        ])

        R = np.eye(3) + np.sin(theta) * A + (1.0 - np.cos(theta)) * (A @ A)

    eigenvecs_sort_traceless = R @ eigvecs_sorted

    # Center NEVPT2 coordinates
    nevpt2_coords_center = nevpt2_coords.mean(axis=0, keepdims=True)
    nevpt2_coords_centerless = nevpt2_coords - nevpt2_coords_center

    # Convert NEVPT2 coordinates to chi frame
    nevpt2_coords_chi_frame = (
        nevpt2_coords_centerless @ eigenvecs_sort_traceless
        + nevpt2_coords_center
    )

    # Clean labels (remove numeric indices, if any)
    clean_labels = [re.sub(r"\d+", "", str(label)) for label in nevpt2_labels]

    # Prepare output directory and filename
    os.makedirs(file_path, exist_ok=True)
    xyz_filename = os.path.join(file_path, "chi_frame_structure.xyz")

    # Build a descriptive comment line
    _comment = (
        f"NEVPT2 coordinates rotated into the susceptibility (chi) frame. "
        f"This file was generated with SimpNMR v{__version__} "
        f"at {datetime.datetime.now().strftime('%H:%M:%S %d-%m-%Y')}."
    )

    # Save XYZ using the helper from xyz_py
    xyzp.save_xyz(
        xyz_filename,
        labels=clean_labels,
        coords=nevpt2_coords_chi_frame,
        verbose=False,
        comment=_comment,
    )

    ut.cprint(f"\n Chi-frame coordinates saved to {xyz_filename}\n", "cyan")

    # Return list of (label, coord) tuples for possible downstream use
    coords_chi_frame_out = list(zip(clean_labels, nevpt2_coords_chi_frame))
    return coords_chi_frame_out